@@ -31,8 +31,7 @@ CONF = cfg.CONF
 LOG = logging.getLogger(__name__)
 
 
-def ensure_base_wiring_config(idl, ovs_idl, ovn_idl=None, routing_tables=None):
-    routing_tables = routing_tables or {}
+def ensure_base_wiring_config(idl, ovs_idl, ovn_idl=None, routing_tables={}):
     if CONF.exposing_method == constants.EXPOSE_METHOD_UNDERLAY:
         return _ensure_base_wiring_config_underlay(idl, ovs_idl,
                                                    routing_tables)
@@ -56,8 +55,6 @@ def _ensure_base_wiring_config_underlay(idl, ovs_idl, routing_tables):
             continue
         ovn_bridge_mappings[network] = bridge
 
-        linux_net.ensure_bridge(bridge)
-
         linux_net.ensure_routing_table_for_bridge(
             routing_tables, bridge, CONF.bgp_vrf_table_id)
         vlan_tags = idl.get_network_vlan_tag_by_network_name(network)
@@ -236,7 +233,16 @@ def _ensure_base_wiring_config_ovn(ovs_idl, ovn_idl):
     _execute_commands(ovn_idl, cmds)
     cmds = []
     cmds.extend(_ensure_ovn_policies(ovn_idl, CONF.local_ovn_cluster.peer_ips))
-    cmds.extend(_ensure_ovn_routes(ovn_idl, CONF.local_ovn_cluster.peer_ips))
+
+    ovs_idl.get_ovn_bridge_mappings(
+        bridge=CONF.local_ovn_cluster.bgp_chassis_id)
+
+    bfds = _ensure_bfds(ovn_idl, zip(
+        CONF.local_ovn_cluster.peer_ips,
+        ovs_idl.get_ovn_bridge_mappings(
+            bridge=CONF.local_ovn_cluster.bgp_chassis_id)))
+    cmds.extend(
+        _ensure_ovn_routes(ovn_idl, CONF.local_ovn_cluster.peer_ips, bfds))
     # Creation of all router related cmds in a single transaction
     _execute_commands(ovn_idl, cmds)
 
@@ -254,18 +260,32 @@ def _ensure_base_wiring_config_ovn(ovs_idl, ovn_idl):
         # and external LS (connecting to the NICs)
         if bridge in ovn_bridge_mappings.values():
             # Internal Bridge connecting to OpenStack OVN cluster
-            _ensure_ovn_network_link(ovn_idl, network, 'internal',
-                                     provider_cidrs=provider_cidrs)
+            _ensure_ovn_network_link_internal(ovn_idl, network, provider_cidrs)
         else:
             ip, mac = linux_net.get_nic_info(bridge)
             # External Bridge connecting to the external networks
-            _ensure_ovn_network_link(ovn_idl, network, 'external',
-                                     ip=ip, mac=mac)
+            _ensure_ovn_network_link_external(ovn_idl, network, ip, mac)
             _ensure_ingress_flows(bridge, mac, network, provider_cidrs)
+            _ensure_fdb_entry(bridge, mac)
 
     return ovn_bridge_mappings, flows_info
 
 
+def _ensure_bfds(ovn_idl, nic_infos):
+    bfds = []
+    for peer_ip, bm in nic_infos:
+        network, bridge = helpers.parse_bridge_mapping(bm)
+        # the port name is hardcoded
+        r_port_name = "{}-{}".format(
+            constants.OVN_CLUSTER_ROUTER, network)
+        bfds.append(
+            ovn_idl.bfd_add(
+                logical_port=r_port_name, dst_ip=peer_ip,
+                may_exist=True).execute(check_error=True))
+
+    return bfds
+
+
 def _ensure_ovn_router(ovn_idl):
     return [ovn_idl.lr_add(constants.OVN_CLUSTER_ROUTER, may_exist=True)]
 
@@ -287,21 +307,58 @@ def _execute_commands(idl, cmds):
             txn.add(command)
 
 
-def _ensure_ovn_network_link(ovn_idl, switch_name, direction,
-                             provider_cidrs=None, ip=None, mac=None):
-    """Base configuration for connecting LR and LSs
+def _ensure_ovn_network_link_internal(ovn_idl, switch_name, provider_cidrs):
+    """Configure connecting LR and LSs internal side
 
-    This function is in charge of connecting the LR to the external or internal
-    LS
-
-    For the internal LS it configures:
     1. Creates LRP to connect to the internal switch
     2. If networks (provider_cidrs) are different, adding the new networks
     3. Create LSP related to the LRP with the right options, including the
        arp_proxy
     4. Bind the LRP to the local chassis
 
-    For the external LS it configures:
+    :param ovn_idl: The idl to communicate with local (in-node) NB DB
+    :param switch_name: the name of the logical switch to configure
+    :param provider_cidrs: CIDRs to configure the networks of the
+                           LRP, as well as to configure the ARP
+                           proxy on the internal LSP
+    """
+    cmds = []
+    # Connect BGP router to the internal logical switch
+    r_port_name = "{}-openstack".format(constants.OVN_CLUSTER_ROUTER)
+    try:
+        ovn_idl.lrp_add(constants.OVN_CLUSTER_ROUTER, r_port_name,
+                        constants.OVN_CLUSTER_ROUTER_INTERNAL_MAC,
+                        provider_cidrs, peer=[], may_exist=True).execute(
+            check_error=True)
+    except RuntimeError as rte:
+        # TODO(ltomasbo): Change OVSDBAPP to return a different error for
+        # this to avoid having to compare strings as this is error prone
+        networks_message = 'with different networks'
+        if networks_message not in str(rte):
+            raise
+        # Trying to sync the networks by adding them
+        cmds.append(ovn_idl.lrp_add_networks(r_port_name, provider_cidrs,
+                                             may_exist=True))
+
+    s_port_name = "openstack-{}".format(constants.OVN_CLUSTER_ROUTER)
+    # NOTE(ltomasbo): require v23.06.0 so that proxy-arp works as expected.
+    # If older version the provider_cidrs should contain all the provider
+    # network cidrs, pointing to the gateway IP of the network.
+    cidrs = ','.join(provider_cidrs) if provider_cidrs else '0.0.0.0/0'
+    options = {'router-port': r_port_name, 'arp_proxy': cidrs}
+    cmds.extend(_ensure_lsp_cmds(ovn_idl, s_port_name, switch_name,
+                                 'router', 'router', **options))
+    # bind to local chassis
+    # ovn-nbctl lrp-set-gateway-chassis  bgp-router-public bgp 1
+    cmds.append(ovn_idl.lrp_set_gateway_chassis(
+        r_port_name, CONF.local_ovn_cluster.bgp_chassis_id, 1))
+
+    _execute_commands(ovn_idl, cmds)
+
+
+def _ensure_ovn_network_link_external(ovn_idl, switch_name, ip, mac):
+    """Configure connecting LR and LSs external side
+
     1. Creates LRP to connect to the external switch
     2. If networks (ip) is different than the nic network add the nic network
        and remove the extra ones
@@ -309,80 +366,45 @@ def _ensure_ovn_network_link(ovn_idl, switch_name, direction,
 
     :param ovn_idl: The idl to communicate with local (in-node) NB DB
     :param switch_name: the name of the logical switch to configure
-    :param direction: can be 'internal' or 'external'
-    :param provider_cidrs (optional): CIDRs to configure the networks of the
-                                      LRP, as well as to configure the ARP
-                                      proxy on the internal LSP
-                                      (only for the internal)
-    :param ip (optional): IP to configure in the LRP connected to the external
-                          switch (only for the external)
-    :param mac (optional): MAC to configure in the LRP connected to the
-                           external switch (only for the external)
+    :param ip: IP to configure in the LRP connected to the external switch
+    :param mac: MAC to configure in the LRP connected to the external switch
     """
-    # It accepts 2 values for direction: internal or external
     cmds = []
-    if direction == 'internal':
-        # Connect BGP router to the internal logical switch
-        r_port_name = "{}-openstack".format(constants.OVN_CLUSTER_ROUTER)
-        try:
-            ovn_idl.lrp_add(constants.OVN_CLUSTER_ROUTER, r_port_name,
-                            constants.OVN_CLUSTER_ROUTER_INTERNAL_MAC,
-                            provider_cidrs, peer=[], may_exist=True).execute(
-                check_error=True)
-        except RuntimeError as rte:
-            # TODO(ltomasbo): Change OVSDBAPP to return a different error for
-            # this to avoid having to compare strings as this is error prone
-            networks_message = 'with different networks'
-            if networks_message not in str(rte):
-                raise
-            # Trying to sync the networks by adding them
-            cmds.append(ovn_idl.lrp_add_networks(r_port_name, provider_cidrs,
-                                                 may_exist=True))
+    r_port_name = "{}-{}".format(constants.OVN_CLUSTER_ROUTER, switch_name)
 
-        s_port_name = "openstack-{}".format(constants.OVN_CLUSTER_ROUTER)
-        # NOTE(ltomasbo): require v23.06.0 so that proxy-arp works as expected.
-        # If older version the provider_cidrs should contain all the provider
-        # network cidrs, pointing to the gateway IP of the network.
-        cidrs = ','.join(provider_cidrs) if provider_cidrs else '0.0.0.0/0'
-        options = {'router-port': r_port_name, 'arp_proxy': cidrs}
-        cmds.extend(_ensure_lsp_cmds(ovn_idl, s_port_name, switch_name,
-                                     'router', 'router', **options))
-        # bind to local chassis
-        # ovn-nbctl lrp-set-gateway-chassis  bgp-router-public bgp 1
-        cmds.append(ovn_idl.lrp_set_gateway_chassis(
-            r_port_name, CONF.local_ovn_cluster.bgp_chassis_id, 1))
-    else:  # direction == 'external'
-        # Connect BGP router to the external logical switch
-        r_port_name = "{}-{}".format(constants.OVN_CLUSTER_ROUTER, switch_name)
-        # LRP
-        try:
-            ovn_idl.lrp_add(constants.OVN_CLUSTER_ROUTER, r_port_name,
-                            mac, [ip], peer=[], may_exist=True).execute(
-                check_error=True)
-        except RuntimeError as rte:
-            # TODO(ltomasbo): Change OVSDBAPP to return a different error for
-            # this to avoid having to compare strings as this is error prone
-            networks_message = 'with different networks'
-            if networks_message not in str(rte):
-                raise
-            # Trying to sync the networks by adding them
-            cmds.append(ovn_idl.lrp_add_networks(r_port_name,
-                                                 [ip],
-                                                 may_exist=True))
-            lrp = ovn_idl.lrp_get(r_port_name).execute(check_error=True)
-            for net in lrp.networks:
-                if net != ip:
-                    cmds.append(ovn_idl.lrp_del_networks(r_port_name,
-                                                         [net],
-                                                         if_exists=True))
-        # LSP
-        s_port_name = "{}-{}".format(switch_name, constants.OVN_CLUSTER_ROUTER)
-        options = {'router-port': r_port_name}
-        cmds.extend(_ensure_lsp_cmds(ovn_idl, s_port_name, switch_name,
-                                     'router', 'router', **options))
+    # LRP
+    try:
+        ovn_idl.lrp_add(constants.OVN_CLUSTER_ROUTER, r_port_name,
+                        mac, [ip], peer=[], may_exist=True).execute(
+            check_error=True)
+    except RuntimeError as rte:
+        # TODO(ltomasbo): Change OVSDBAPP to return a different error for
+        # this to avoid having to compare strings as this is error prone
+        networks_message = 'with different networks'
+        if networks_message not in str(rte):
+            raise
+        # Trying to sync the networks by adding them
+        cmds.append(ovn_idl.lrp_add_networks(r_port_name,
+                                             [ip],
+                                             may_exist=True))
+        lrp = ovn_idl.lrp_get(r_port_name).execute(check_error=True)
+        for net in lrp.networks:
+            if net != ip:
+                cmds.append(ovn_idl.lrp_del_networks(r_port_name,
+                                                     [net],
+                                                     if_exists=True))
+    # LSP
+    s_port_name = "{}-{}".format(switch_name, constants.OVN_CLUSTER_ROUTER)
+    options = {'router-port': r_port_name}
+    cmds.extend(_ensure_lsp_cmds(ovn_idl, s_port_name, switch_name,
+                                 'router', 'router', **options))
 
-    if cmds:
-        _execute_commands(ovn_idl, cmds)
+    # bind to local chassis
+    # ovn-nbctl lrp-set-gateway-chassis  bgp-router-public bgp 1
+    cmds.append(ovn_idl.lrp_set_gateway_chassis(
+        r_port_name, CONF.local_ovn_cluster.bgp_chassis_id, 1))
+
+    _execute_commands(ovn_idl, cmds)
 
 
 def _ensure_lsp_cmds(ovn_idl, port_name, switch, port_type, addresses,
@@ -410,12 +432,13 @@ def _ensure_ovn_policies(ovn_idl, next_hops):
                                   match, action, may_exist=True, **columns)]
 
 
-def _ensure_ovn_routes(ovn_idl, peer_ips):
+def _ensure_ovn_routes(ovn_idl, peer_ips, bfds):
     prefix = '0.0.0.0/0'
     cmds = []
-    for ip in peer_ips:
+    for bfd, ip in zip(bfds, peer_ips):
         cmds.append(ovn_idl.lr_route_add(constants.OVN_CLUSTER_ROUTER, prefix,
-                                         ip, ecmp=True, may_exist=True))
+                                         ip, ecmp=True, may_exist=True,
+                                         bfd=bfd.uuid))
     return cmds
 
 
@@ -460,6 +483,18 @@ def _ensure_ingress_flows(bridge, mac, switch_name, provider_cidrs):
         ovs.ensure_flow(bridge, flow)
 
 
+def _ensure_fdb_entry(bridge, mac):
+    """Ensure static FDB entries for internal ports of a fabric bridge.
+
+    Fabric-facing bridges need to have a static MAC entry to avoid dynamic
+    re-learning in their FDB as this will lead to drop flows being generated
+    by OVS due to flow table misses.
+    """
+    # The bridge name is also the name of its internal port. The MAC address
+    # is the MAC address of the bridge itself.
+    ovs.add_fdb_entry(bridge, bridge, mac, 0)
+
+
 def _ensure_egress_flows(bridge, patch_ports):
     # outcomming traffic flows
     # patch=`ovs-ofctl show br-provider | grep patch | grep provnet |
@@ -490,19 +525,37 @@ def cleanup_wiring(idl, bridge_mappings, ovs_flows, exposed_ips,
     elif CONF.exposing_method == constants.EXPOSE_METHOD_VRF:
         return _cleanup_wiring_evpn(ovs_flows, routing_tables_routes)
     elif CONF.exposing_method == constants.EXPOSE_METHOD_OVN:
-        # TODO(ltomasbo): clean up old policies, routes and proxy_arps cidrs
-        return True
+        return _cleanup_wiring_ovn(exposed_ips)
+
+
+def _cleanup_extra_ips(exposed_ips: list[str]) -> set[str]:
+    current_ips = set(linux_net.get_exposed_ips(CONF.bgp_nic))
+    expected_ips = {ip for ip_dict in exposed_ips.values()
+                    for ip in ip_dict.keys()}
+    ips_to_delete = current_ips - expected_ips
+    if ips_to_delete:
+        LOG.warning(f"Removing IPs {ips_to_delete} from {CONF.bgp_nic} as "
+                    "they are not expected to be exposed.")
+        linux_net.delete_exposed_ips(ips_to_delete, CONF.bgp_nic)
+    return expected_ips
+
+
+def _cleanup_wiring_ovn(exposed_ips):
+    """Cleanup IPs that are not expected to be exposed on the bgp_nic.
+
+    This may be needed in case ovn-bgp-agent misses updates from the OVN NB
+    database due to temporary downtime, networking issues between the
+    ovn-bgp-agent and the NB DB or otherwise.
+    """
+    _cleanup_extra_ips(exposed_ips)
+    # TODO(dmitriis): clean up other state: in the local OVN NB: logical
+    # router policies, logical router routes, ARP proxies in logical switch
+    # ports, static mac bindings.
 
 
 def _cleanup_wiring_underlay(idl, bridge_mappings, ovs_flows, exposed_ips,
                              routing_tables, routing_tables_routes):
-    current_ips = linux_net.get_exposed_ips(CONF.bgp_nic)
-    expected_ips = [ip for ip_dict in exposed_ips.values()
-                    for ip in ip_dict.keys()]
-
-    ips_to_delete = [ip for ip in current_ips if ip not in expected_ips]
-    linux_net.delete_exposed_ips(ips_to_delete, CONF.bgp_nic)
-
+    expected_ips = _cleanup_extra_ips(exposed_ips)
     extra_routes = {}
     for bridge in bridge_mappings.values():
         extra_routes[bridge] = (
